from __future__ import annotations

from typing import Final, Optional

from attrs import define, field
from textx.exceptions import TextXSemanticError

from tsdoc0.python.answer_shape import AnswerShape
from tsdoc0.python.error_type import ErrorType
from tsdoc0.python.model import Model
from tsdoc0.python.segment import Segment
from tsdoc0.python.solution_code import SolutionCode
from tsdoc0.python.solution_shape import SolutionShape
from tsdoc0.python.utils import repr_parent


@define(auto_attribs=True, kw_only=True)
class QuestionShape(Segment):
    parent: Optional[Model] = field(eq=False, repr=repr_parent)
    indentation: Final[str]
    question: Final[str]
    answer: Optional[AnswerShape]

    @classmethod
    def process_optional_answer(cls, question_shape: QuestionShape) -> None:
        if question_shape.answer:
            return

        if question_shape.parent is None:
            return

        segments = question_shape.parent.segments
        index_next = segments.index(question_shape) + 1
        segments_next = segments[index_next:]

        for segment in segments_next:
            if type(segment) is SolutionShape:
                question_shape.answer = AnswerShape(
                    parent=question_shape,
                    indentation=question_shape.indentation,
                    text=segment.answer,
                )
                return

            if type(segment) is SolutionCode:
                break

        raise TextXSemanticError(
            "Found no answer for a Shape Question",
            line=question_shape.line,
            col=question_shape.col,
            filename=question_shape.filename,
            err_type=ErrorType.FOUND_NO_ANSWER,
        )



from typing import Final, Iterable, Optional

from attrs import define, field

from tsdoc0.python.model import Model
from tsdoc0.python.segment import Segment
from tsdoc0.python.utils import repr_parent
from tsdoc0.python.vocabulary_term import VocabularyTerm


# This converter wrapper-function is used because of a bug with the mypy-attrs plugin.
# https://github.com/python/mypy/issues/8389
def _tuple(iterable: Iterable[VocabularyTerm]) -> tuple[VocabularyTerm, ...]:
    return tuple(iterable)


@define(auto_attribs=True, kw_only=True)
class Vocabulary(Segment):
    parent: Optional[Model] = field(eq=False, repr=repr_parent)
    indentation: Final[str]
    terms: Final[tuple[VocabularyTerm, ...]] = field(converter=_tuple)



from typing import Final, Optional

from attrs import define, field

from tsdoc0.python.activity_part import ActivityPart
from tsdoc0.python.model import Model
from tsdoc0.python.segment import Segment
from tsdoc0.python.utils import repr_parent


@define(auto_attribs=True, kw_only=True)
class ActivityHeader(Segment):
    parent: Optional[Model] = field(eq=False, repr=repr_parent)
    subject_name: Final[str]
    activity_type: Final[str]
    topic_name: Final[str]
    part: Optional[ActivityPart]
    program_name: Final[str]

    @property
    def code(self) -> str:
        return (
            '"""\n'
            f"{self.subject_name}\n"
            f"{self.activity_type}: {self.topic_name}\n"
            f"{self.part.code if self.part else ''}{self.program_name}\n"
            '"""'
        )



from textwrap import fill
from typing import TYPE_CHECKING, Final, Union

from attrs import define, field

from tsdoc0.python.segment import Segment
from tsdoc0.python.utils import repr_parent

# Conditional import used because there would be a circular dependency
# https://mypy.readthedocs.io/en/latest/common_issues.html#import-cycles
if TYPE_CHECKING:
    from tsdoc0.python.comment_abstract import CommentAbstract
    from tsdoc0.python.comment_concrete import CommentConcrete


@define(auto_attribs=True, kw_only=True)
class CommentSentence(Segment):
    # Forward reference used because there would be a circular dependency
    # https://mypy.readthedocs.io/en/latest/kinds_of_types.html#class-name-forward-references
    parent: Union["CommentAbstract", "CommentConcrete", None] = field(
        eq=False, repr=repr_parent
    )
    indentation: Final[str]
    text: Final[str]

    @property
    def code(self) -> str:
        text_pruned = self.text.replace("`", "").removesuffix(".")
        return self._wrap_comment_or_ignore_string_literal(text_pruned)

    def _wrap_comment_or_ignore_string_literal(self, text: str) -> str:
        # Entire-line string literals are not wrapped because students' may want to
        # copy-and-paste them into their code.
        if text.startswith('"') and text.endswith('"'):
            return f"{self.indentation}# {text}"

        indent = f"{self.indentation}# "
        width = 55

        return fill(text, width, initial_indent=indent, subsequent_indent=indent)

    @property
    def callout(self) -> str:
        return self.text



from typing import TYPE_CHECKING, Final, Optional

from attrs import define, field

from tsdoc0.python.segment import Segment
from tsdoc0.python.utils import repr_parent

# Conditional import used because there would be a circular dependency
# https://mypy.readthedocs.io/en/latest/common_issues.html#import-cycles
if TYPE_CHECKING:
    from tsdoc0.python.question_abstract import QuestionAbstract  # noqa: F401


@define(auto_attribs=True, kw_only=True)
class AnswerAbstract(Segment):
    # Forward reference used because there would be a circular dependency
    # https://mypy.readthedocs.io/en/latest/kinds_of_types.html#class-name-forward-references
    parent: Optional["QuestionAbstract"] = field(eq=False, repr=repr_parent)
    indentation: Final[str]
    text: Final[str]



from typing import Final, Iterable

from attrs import define, field

from tsdoc0.python.segment import Segment


# This converter wrapper-function is used because of a bug with the mypy-attrs plugin.
# https://github.com/python/mypy/issues/8389
def _tuple(iterable: Iterable[Segment]) -> tuple[Segment, ...]:
    return tuple(iterable)


# This class must be a non-slotted because there are several special attributes on
# model objects created by textX. All special attributes' names start with prefix _tx.
# https://textx.github.io/textX/3.1/model/#special-model-objects-attributes
@define(auto_attribs=True, kw_only=True, slots=False)
class Model:
    segments: Final[tuple[Segment, ...]] = field(converter=_tuple)



from typing import Final, Optional

from attrs import define, field

from tsdoc0.python.model import Model
from tsdoc0.python.segment import Segment
from tsdoc0.python.utils import repr_parent


@define(auto_attribs=True, kw_only=True)
class SolutionCode(Segment):
    parent: Optional[Model] = field(eq=False, repr=repr_parent)
    indentation: Final[str]
    text: Final[str]

    @property
    def code(self) -> str:
        return f"{self.indentation}{self.text}"

    @property
    def callout(self) -> str:
        return self.code



from dataclasses import dataclass
from typing import Optional, cast

from textx.model import get_model


class Segment:
    @dataclass
    class LineCol:
        line: int
        col: int

    @property
    def linecol(self) -> Optional[LineCol]:
        model = get_model(self)

        if not hasattr(model, "_tx_parser"):
            return None

        tx_parser = model._tx_parser

        # Silence the error about a missing attribute because it already exists inside
        # of all textX model classes
        tx_position = self._tx_position  # type: ignore[attr-defined]

        return Segment.LineCol(*tx_parser.pos_to_linecol(tx_position))

    @property
    def line(self) -> Optional[int]:
        if (linecol := self.linecol) is None:
            return None

        return linecol.line

    @property
    def col(self) -> Optional[int]:
        if (linecol := self.linecol) is None:
            return None

        return linecol.col

    @property
    def filename(self) -> Optional[str]:
        model = get_model(self)

        if not hasattr(model, "_tx_filename"):
            return None

        return cast(str, model._tx_filename)



from typing import Final, Optional

from attr import define, field

from tsdoc0.python.model import Model
from tsdoc0.python.segment import Segment
from tsdoc0.python.utils import repr_parent


@define(auto_attribs=True, kw_only=True)
class PointTricky(Segment):
    parent: Optional[Model] = field(eq=False, repr=repr_parent)
    indentation: Final[str]
    point: Final[str]
    explanation: Final[str]



from typing import Final, Iterable, Optional

from attrs import define, field

from tsdoc0.python.comment_sentence import CommentSentence
from tsdoc0.python.model import Model
from tsdoc0.python.segment import Segment
from tsdoc0.python.utils import repr_parent


# This converter wrapper-function is used because of a bug with the mypy-attrs plugin.
# https://github.com/python/mypy/issues/8389
def _tuple(iterable: Iterable[CommentSentence]) -> tuple[CommentSentence, ...]:
    return tuple(iterable)


@define(auto_attribs=True, kw_only=True)
class CommentAbstract(Segment):
    parent: Optional[Model] = field(eq=False, repr=repr_parent)
    indentation: Final[str]
    sentences: Final[tuple[CommentSentence, ...]] = field(converter=_tuple)

    @property
    def answer(self) -> str:
        return self.sentences[0].callout if len(self.sentences) else ""

    @property
    def code(self) -> str:
        return "\n".join(sentence.code for sentence in self.sentences)

    @property
    def callout(self) -> str:
        return "\n".join(sentence.callout for sentence in self.sentences)



from csv import DictReader
from pathlib import Path
from typing import TYPE_CHECKING, Final, Optional

from attrs import define, field

from tsdoc0.python.segment import Segment
from tsdoc0.python.utils import repr_parent

# Conditional import used because there would be a circular dependency
# https://mypy.readthedocs.io/en/latest/common_issues.html#import-cycles
if TYPE_CHECKING:
    from tsdoc0.python.vocabulary import Vocabulary  # noqa: F401

with open(Path(__file__, "..", "vocabulary_terms.csv").resolve()) as csvfile:
    _DICTIONARY: Final[tuple[dict[str, str], ...]] = tuple(
        row for row in DictReader(csvfile, strict=True)
    )

_NAMES: Final[dict[str, str]] = {
    row["text"]: row["name"] for row in _DICTIONARY if len(row["name"]) > 0
}

_DEFINITIONS: Final[dict[str, str]] = {
    row["text"]: row["definition"] for row in _DICTIONARY
}

_INSTRUCTIONS: Final[dict[str, str]] = {
    row["text"]: row["instruction"] for row in _DICTIONARY
}


@define(auto_attribs=True, kw_only=True)
class VocabularyTerm(Segment):
    # Forward reference used because there would be a circular dependency
    # https://mypy.readthedocs.io/en/latest/kinds_of_types.html#class-name-forward-references
    parent: Optional["Vocabulary"] = field(eq=False, repr=repr_parent)
    text: Final[str]

    @property
    def name(self) -> str:
        return _NAMES.get(self.text, self.text)

    @property
    def definition(self) -> str:
        return _DEFINITIONS[self.text]

    @property
    def instruction(self) -> str:
        return _INSTRUCTIONS[self.text]



from typing import Final, Iterable, Optional

from attrs import define, field

from tsdoc0.python.model import Model
from tsdoc0.python.segment import Segment
from tsdoc0.python.solution_item import SolutionItem
from tsdoc0.python.solution_pseudocode import SolutionPseudocode
from tsdoc0.python.utils import repr_parent


# This converter wrapper-function is used because of a bug with the mypy-attrs plugin.
# https://github.com/python/mypy/issues/8389
#
# The Segment type is used instead of a generic type because of another bug with the
# mypy-attrs plugin.
# https://github.com/python/mypy/issues/8625#issuecomment-632720205
def _tuple(iterable: Iterable[Segment]) -> tuple[Segment, ...]:
    return tuple(iterable)


@define(auto_attribs=True, kw_only=True)
class SolutionShape(Segment):
    parent: Optional[Model] = field(eq=False, repr=repr_parent)
    indentation: Final[str]
    pseudocodes: Final[tuple[SolutionPseudocode, ...]] = field(converter=_tuple)
    items: Final[tuple[SolutionItem, ...]] = field(converter=_tuple)

    @property
    def answer(self) -> str:
        pseudocodes = "\n".join(pseudocode.text for pseudocode in self.pseudocodes)
        items = "\n".join(f"{item.key}: {item.value}" for item in self.items)

        return f"```\n{pseudocodes}\n\n{items}\n```"

    @property
    def callout(self) -> str:
        pseudocodes = "\n".join(pseudocode.text for pseudocode in self.pseudocodes)
        items = "\n".join(f"{item.key}: `{item.value}`" for item in self.items)

        return f"```\n{pseudocodes}\n```\n\n{items}"



from __future__ import annotations

from typing import Final, Optional

from attrs import define, field
from textx.exceptions import TextXSemanticError

from tsdoc0.python.answer_abstract import AnswerAbstract
from tsdoc0.python.comment_abstract import CommentAbstract
from tsdoc0.python.error_type import ErrorType
from tsdoc0.python.model import Model
from tsdoc0.python.segment import Segment
from tsdoc0.python.solution_code import SolutionCode
from tsdoc0.python.utils import repr_parent


@define(auto_attribs=True, kw_only=True)
class QuestionAbstract(Segment):
    parent: Optional[Model] = field(eq=False, repr=repr_parent)
    indentation: Final[str]
    question: Final[str]
    answer: Optional[AnswerAbstract]

    @classmethod
    def process_optional_answer(cls, question_abstract: QuestionAbstract) -> None:
        if question_abstract.answer:
            return

        if question_abstract.parent is None:
            return

        segments = question_abstract.parent.segments
        index_next = segments.index(question_abstract) + 1
        segments_next = segments[index_next:]

        for segment in segments_next:
            if type(segment) is CommentAbstract:
                question_abstract.answer = AnswerAbstract(
                    parent=question_abstract,
                    indentation=question_abstract.indentation,
                    text=segment.answer,
                )
                return

            if type(segment) is SolutionCode:
                break

        print(QuestionAbstract.__slots__)

        raise TextXSemanticError(
            "Found no answer for an Abstract Question",
            line=question_abstract.line,
            col=question_abstract.col,
            filename=question_abstract.filename,
            err_type=ErrorType.FOUND_NO_ANSWER,
        )



from os.path import dirname, join

from textx import language, metamodel_from_file
from textx.metamodel import TextXMetaModel

from tsdoc0 import __version__
from tsdoc0.python.activity_header import ActivityHeader
from tsdoc0.python.activity_part import ActivityPart
from tsdoc0.python.answer_abstract import AnswerAbstract
from tsdoc0.python.answer_shape import AnswerShape
from tsdoc0.python.code_refactorable import CodeRefactorable
from tsdoc0.python.code_static import CodeStatic
from tsdoc0.python.comment_abstract import CommentAbstract
from tsdoc0.python.comment_concrete import CommentConcrete
from tsdoc0.python.comment_paragraph import CommentParagraph
from tsdoc0.python.comment_run_your_code import CommentRunYourCode
from tsdoc0.python.comment_section import CommentSection
from tsdoc0.python.comment_sentence import CommentSentence
from tsdoc0.python.comment_turn_in_your_code import CommentTurnInYourCode
from tsdoc0.python.error_type import ErrorType  # noqa :F401
from tsdoc0.python.keywords import Keywords
from tsdoc0.python.line_blank import LineBlank
from tsdoc0.python.model import Model
from tsdoc0.python.point_tricky import PointTricky
from tsdoc0.python.question_abstract import QuestionAbstract
from tsdoc0.python.question_review import QuestionReview
from tsdoc0.python.question_shape import QuestionShape
from tsdoc0.python.question_structure import QuestionStructure
from tsdoc0.python.question_structure_no import QuestionStructureNo
from tsdoc0.python.segment import Segment  # noqa :F401
from tsdoc0.python.solution_code import SolutionCode
from tsdoc0.python.solution_item import SolutionItem
from tsdoc0.python.solution_pseudocode import SolutionPseudocode
from tsdoc0.python.solution_shape import SolutionShape
from tsdoc0.python.vocabulary import Vocabulary
from tsdoc0.python.vocabulary_term import VocabularyTerm

filename_ext = f".tsdoc{__version__.split('.')[0]}.py"

metamodel: TextXMetaModel = metamodel_from_file(
    join(dirname(__file__), "grammar.tx"),
    classes=[
        ActivityHeader,
        ActivityPart,
        AnswerAbstract,
        AnswerShape,
        CodeRefactorable,
        CodeStatic,
        CommentAbstract,
        CommentConcrete,
        CommentParagraph,
        CommentRunYourCode,
        CommentSection,
        CommentSentence,
        CommentTurnInYourCode,
        Keywords,
        Model,
        SolutionCode,
        LineBlank,
        PointTricky,
        QuestionAbstract,
        QuestionReview,
        QuestionShape,
        QuestionStructure,
        QuestionStructureNo,
        SolutionItem,
        SolutionPseudocode,
        SolutionShape,
        VocabularyTerm,
        Vocabulary,
    ],
    use_regexp_group=True,
)

metamodel.register_obj_processors(
    {
        "QuestionAbstract": QuestionAbstract.process_optional_answer,
        "QuestionShape": QuestionShape.process_optional_answer,
    }
)


@language("TSDoc Python", f"*{filename_ext}")
def register() -> TextXMetaModel:
    "An embedded comment language for TechSmart Python files."
    return metamodel
